#include <cassert>

#include <cstddef>
using std::size_t;

#include <cstdlib>
using std::exit;
using std::atoi;

#include <cmath>
using std::sqrt;
using std::fabs;
using std::fmod;
using std::pow;

#include <vector>
using std::vector;

#include <set>
using std::set;

#include <algorithm>
using std::max;
using std::min;

#include <iostream>
using std::ostream;
using std::cout;
using std::cerr;
using std::endl;

#include <iomanip>
using std::setw;
using std::setprecision;

#include <fstream>
using std::ofstream;

#include <sstream>
using std::stringstream;

#ifdef FORTRAN_TRAILING_UNDERSCORE
#define F77NAME(x) x##_
#else
#define F77NAME(x) x
#endif

#ifdef __APPLE__
#include <Accelerate/Accelerate.h>
//#include <vecLib/clapack.h>
typedef __CLPK_doublereal doublereal;
typedef __CLPK_integer integer;
#else
typedef double doublereal;
typedef int integer;
// or typedef long int integer; on some systems
extern "C" void F77NAME(dspevd)
(char*, char*, integer*, doublereal*, doublereal*,
 doublereal*, integer*, doublereal*, integer*,
 integer*, integer*, integer*);
#endif

const size_t L = 20;
const size_t N = 2*L*L;
const size_t fermi_level = N/2;
const size_t z = 3;
const double xscale = sqrt(3.0);

const bool open_bcs_WE = false;
const bool open_bcs_NS = false;

class pairing
{
public:
	size_t i;
	size_t j;
	pairing() : i(0), j(0) {}
	pairing(size_t i_, size_t j_) : i(i_), j(j_) {}
};

class site
{
public:
	double x;
	double y;
	site() : x(0), y(0) {}
	site(double x_, double y_) : x(x_), y(y_) {}
};

class site2
{
public:
	size_t i;
	double x;
	double y;
	site2() : i(0), x(0), y(0) {}
	site2(size_t i_, double x_, double y_) : i(i_), x(x_), y(y_) {}
	friend bool operator<(const site2 &p, const site2 &q) { return p.y < q.y or (fabs(p.y - q.y) < 1E-6 and p.x < q.x); }
};

const bool has_edge_states = open_bcs_WE and open_bcs_NS;

inline double inner_product(const site &r1, const site &r2)
{
    return r1.x*r2.x + r1.y*r2.y;
}

bool positions_ready = false;
inline double distance(const site &r1, const site &r2)
{
	assert(positions_ready);
	const double dx = r1.x-r2.x;
	const double dy = r1.y-r2.y;
    return sqrt(dx*dx + dy*dy);
}

inline double distance_x(const site &r1, const site &r2)
{
	assert(positions_ready);
    return fabs(r1.x-r2.x);
}

inline double distance_y(const site &r1, const site &r2)
{
	assert(positions_ready);
    return fabs(r1.y-r2.y);
}

inline double distance2(const site &r1, const site &r2)
{
	assert(positions_ready);
	const double dx = r1.x-r2.x;
	const double dy = r1.y-r2.y;
    return dx*dx + dy*dy;
}

vector<pairing> hopping_list;
vector<site> site_list(N);
set<site2> site_set;

inline double distance(size_t i, size_t j)
{
    return distance(site_list[i],site_list[j]);
}

inline double distance_x(size_t i, size_t j)
{
    return distance_x(site_list[i],site_list[j]);
}

inline double distance_y(size_t i, size_t j)
{
    return distance_y(site_list[i],site_list[j]);
}

inline double distance2(size_t i, size_t j)
{
    return distance2(site_list[i],site_list[j]);
}

enum edge_type { INTERIOR = 0x00, LEFT = 0x01, RIGHT = 0x02, BOTTOM = 0x04, TOP = 0x08 };
class site_attributes
{
public:
    size_t coord;
    size_t neighbour[3];
	char sublattice;
};
vector<site_attributes> lattice_info(N);

inline void connect(size_t i1, size_t i2, double t)
{
    if (i1 < i2)
		hopping_list.push_back( pairing(i1,i2) );
}

void build_honeycomb_lattice(void)
{
	// sqrt(3)*L x 1.5*L 
	// determine nearest-neighbour connectivity
	size_t i = 0;
	for (size_t y = 0; y < 2*L; ++y)
	{
		for (size_t x = 0; x < L; ++x)
		{
			unsigned int edge = 0;
			if (x == 0 and y%2 == 1)
				edge |= LEFT;
			else if (x == L-1 and y%2 == 0)
				edge |= RIGHT;
			if (y == 0)
				edge |= BOTTOM;
			else if (y == 2*L-1)
				edge |= TOP;
			if (y%2 == 0)
			{
				// up-right;
				if (edge & RIGHT)
					lattice_info[i].neighbour[0] = i+1;
				else
					lattice_info[i].neighbour[0] = i+L+1;
				// up-left
				lattice_info[i].neighbour[1] = i+L;
				// down;
				if (edge & BOTTOM)
					lattice_info[i].neighbour[2] = N-L+(i+L/2)%L;
				else
					lattice_info[i].neighbour[2] = i-L;
			}
			else
			{
				// down-right;
				lattice_info[i].neighbour[0] = i-L;
				// down-left
				if (edge & LEFT)
					lattice_info[i].neighbour[1] = i-1;
				else
					lattice_info[i].neighbour[1] = i-L-1;
				// up;
				if (edge & TOP)
					lattice_info[i].neighbour[2] = ((i+L)%N+L/2)%L;
				else
					lattice_info[i].neighbour[2] = i+L;
			}
			lattice_info[i].coord = 3;
			lattice_info[i].sublattice = (y%2 == 0 ? '0' : '1');
			++i;
		} // for (x)
	} // for (y)
	
	// compute lattice site positions
	double xoffset = 0.0;
	double yoffset = 0.0;
	i=0;
	for (size_t y = 0; y < 2*L; ++y)
	{
		if ((y+1)%2 == 0)
		{
			xoffset -= 0.5;
			yoffset -= 0.5;
		}
		
		for (size_t x = 0; x < L; ++x)
		{
			const double xx = xscale*(fmod(L+x+xoffset,L));
			const double yy = y+yoffset;
			site_set.insert(site2(i,xx,yy));
			site_list[i++] = site(xx,yy);
		}
	}
	assert(site_list.size() == site_set.size());
	positions_ready = true;
	
	i=0;
	for (size_t y = 0; y < 2*L; ++y)
	{
		for (size_t x = 0; x < L; ++x)
		{
			if ( lattice_info[i].coord > 0)
			{
				if ( (distance_x(i,lattice_info[i].neighbour[0]) < 2.0 or !open_bcs_WE) and
					 (distance_y(i,lattice_info[i].neighbour[0]) < 2.0 or !open_bcs_NS) )
					connect(i,lattice_info[i].neighbour[0],1.0);
				if ( (distance_x(i,lattice_info[i].neighbour[1]) < 2.0 or !open_bcs_WE) and
					(distance_y(i,lattice_info[i].neighbour[1]) < 2.0 or !open_bcs_NS) )
					connect(i,lattice_info[i].neighbour[1],1.0);
				if ( (distance_x(i,lattice_info[i].neighbour[2]) < 2.0 or !open_bcs_WE) and
					(distance_y(i,lattice_info[i].neighbour[2]) < 2.0 or !open_bcs_NS) )
					connect(i,lattice_info[i].neighbour[2],1.0);
			}
			++i;
		} // for (x)
	} // for (y)
	
	assert(hopping_list.size() <= z*N/2);
	assert(open_bcs_WE or open_bcs_NS or hopping_list.size() == z*N/2);	
}

void ps_draw_hopping(size_t i, size_t j, ofstream &fout)
{
	const double dx = site_list[j].x-site_list[i].x;
	const double dy = site_list[j].y-site_list[i].y;
	if (fabs(dx) < 2 and fabs(dy) < 2)
	{
		fout << "newpath" << endl;
		fout << site_list[i].x << " " << site_list[i].y << " moveto" << endl;
		fout << site_list[j].x << " " << site_list[j].y << " lineto" << endl;
		fout << "closepath\nstroke" << endl;
	}
	else
	{
		const double dxx = (dx > 2 ? dx-xscale*L : ( dx < -2 ? dx+xscale*L : dx ) );
		const double dyy = (dy > 2 ? dy-1.5*L : ( dy < -2 ? dy+1.5*L : dy ) );
		fout << "newpath" << endl;
		fout << site_list[i].x << " " << site_list[i].y << " moveto\n";
		fout << site_list[i].x + 0.5*dxx << " " << site_list[i].y + 0.5*dyy << " lineto\nclosepath\nstroke\nnewpath" << endl;
		fout << site_list[j].x << " " << site_list[j].y << " moveto" << endl;
		fout << site_list[j].x - 0.5*dxx << " " << site_list[j].y - 0.5*dyy << " lineto\nclosepath\nstroke" << endl;
	}	
}

void ps_draw_lattice(void)
{
	ofstream fout("config.eps");
    fout << "%!PS-Adobe-3.0 EPSF-3.0" << endl;
    fout << "%%BoundingBox: 0 0 612 530" << endl;
    const double s = 612.0/sqrt(3.0)/(L+0.5);
    fout << s << " " << s << " scale" << endl;
    fout << "0.86 0.86 translate" << endl;
	
	fout << "1 setlinecap" << endl;
	fout << "0.04 setlinewidth" << endl;
	fout << "0.0 setgray" << endl;
	
	for (vector<pairing>::const_iterator step = hopping_list.begin(); step != hopping_list.end(); ++step)
		ps_draw_hopping(step->i,step->j,fout);
	
	fout << "0.05 setlinewidth" << endl;
	fout << "0 setgray" << endl;
	for (size_t i = 0; i < site_list.size(); ++i)
	{
		const double radius = 0.15;
		const double shade = (lattice_info[i].sublattice == '0' ? 0.66 : 1.0);
		fout << "newpath\n";
		fout << site_list[i].x << " " << site_list[i].y << " " << radius << " 0 360 arc" << endl;
		fout << "closepath\n"
		"gsave\n"
		<< shade << " setgray\n"
		"fill\n"
		"grestore\n"
		"stroke" << endl;
	}
	fout.close();	
}

bool eigensolve(vector<doublereal> &H, vector<doublereal> &Eval, vector<doublereal> &Evec)
{
	// Solve the eigenvalue problem with LAPACK's dsepvd routine

	assert(H.size() == N*(N+1)/2);
	assert(Eval.size() == N);
	assert(Evec.size() == N*N);
	
	integer info = 1;
	
	// (!) add code for question 2
	//     that puts the eigenvalues
	//     and eigenvectors of H into
	//     Eval and Evecs

	return (info == 0);
}

void write_mag(vector<double> &Psireal, vector<double> &Psiimag, ofstream &fgp)
{
	static unsigned int dat_count = 0;
	stringstream ss;
	ss << "g-";
	ss.fill('0');
	ss.width(5);
	ss << dat_count++ << ".dat";
	if (dat_count == 1)
		fgp << "set zrange [] writeback" << endl;	
	fgp << "splot \"" << ss.str() << "\" with lines" << endl;
	if (dat_count == 1)
		fgp << "pause -1 \"Press return to start\"" << endl;
	else
	{
		fgp << "set zrange restore" << endl;
		fgp << "pause 0.05" << endl;
	}
	ofstream fout(ss.str().c_str());
	
	double lasty = -1.0;
	for (set<site2>::const_iterator step = site_set.begin(); step != site_set.end(); ++step)
	{
		const size_t i = step->i;
		const double x = step->x;
		const double y = step->y;
		if (y != lasty)
		{
			lasty = y;
			fout << endl;
		}
		fout << x << " " << y << " " << (Psireal[i]*Psireal[i] + Psiimag[i]*Psiimag[i]) << endl;
	}		
		
	fout.close();	
}

void usage(void)
{
	cerr << "usage: graphene (question# = 1, 2, 3, or 4)" << endl;
	exit(1);
}

int main(int argc, char* argv[])
{
	build_honeycomb_lattice();
	ps_draw_lattice();

   if (argc == 1)
      exit(0);
	if (argc != 2)
		usage();
	int question = atoi(argv[1]);
	if (question < 1 or question > 4)
		usage();
	
	// set up storage for the Hamiltonian, Eigenvalues, and Eigenvectors
	vector<doublereal> H(N*(N+1)/2);
	vector<doublereal> Eval(N);
	vector<doublereal> Evec(N*N);
	
	// initialize Hamiltonian to zero
	for (size_t n = 0; n < H.size(); ++n)
		H[n] = 0.0;
	
	// (!) add code for question 1 
	//     that sets all nearest neighbour 
	//     connections in H
	
	if (question == 1)
		exit(0);
	
	const bool success = eigensolve(H,Eval,Evec);
		
	if (!success) 
	{
		cerr << "There's a problem with the eigensolver" << endl;
		exit(1);
	}

	if (question == 2)
	{
#ifndef NDEBUG	
		// check that eigenvectors are stored in columns and normalized
		for (size_t n = 0; n < N; ++n)
		{
			double sum = 0.0;
			for (size_t i = 0; i < N; ++i)
				sum += Evec[n+i*N]*Evec[n+i*N];
			if (fabs(sum-1) > 1E-5) 
			{
				cerr << "There's a problem with the eigenvectors" << endl;
				exit(1);
			}
		}
#endif // NDEBUG
		exit(0);
	}

	// (!) add code for question 3
	//     that computes the density
	//     of states, preferably dumping 
	//     binned and normalized energy 
	//     data to a file 'dos.dat'
	
	if (question == 3)
		exit(0);
	
	vector<double> wfreal(N);
	vector<double> wfimag(N);
	vector<double> cR(N);
	vector<double> cI(N);

	const double K1x = 4.0*M_PI/(3.0*sqrt(3.0));
	const double K1y = 0.0;
	
	const double K2x = 2.0*M_PI/(3.0*sqrt(3.0));
	const double K2y = 2.0*M_PI/3.0;
	
	const size_t i0 = L*L+L/2+1;
	//const double kx = -0.6*K1x;
	//const double ky = 0.15*K1x;
	const double kx = 0.0;
	const double ky = 0.0;
	const double gammak = 2.0*cos(sqrt(3.0)*kx) + 4.0*cos(0.5*sqrt(3.0)*kx)*cos(1.5*ky);
	const double Ek = sqrt(3.0+gammak);
	
	double sum = 0.0;
	for (size_t alpha = 0; alpha < N; ++alpha)
		cR[alpha] = cI[alpha] = 0.0;

	for (size_t alpha = fermi_level; alpha < N; ++alpha)
	{
		for (size_t i = 0; i < N; ++i)
		{
			const double aa = 0.1;
			const double vi = Evec[i+N*alpha]*exp(-aa*distance2(i,i0));
			const double x = site_list[i].x;
			const double y = site_list[i].y;
			if (lattice_info[i].sublattice == '0')
			{
				const double phi = kx*x + ky*y;
				cR[alpha] += vi*cos(phi);
				cI[alpha] += vi*sin(phi);
			}
			else
			{
				const double zR = -( cos(kx*(x+0.5*sqrt(3.0))+ky*(y+0.5)) + cos(kx*x+ky*(y-1.0)) + cos(kx*(x-0.5*sqrt(3.0))+ky*(y+0.5)) )/Ek;
				const double zI = -( sin(kx*(x+0.5*sqrt(3.0))+ky*(y+0.5)) + sin(kx*x+ky*(y-1.0)) + sin(kx*(x-0.5*sqrt(3.0))+ky*(y+0.5)) )/Ek;
				cR[alpha] += vi*zR;
				cI[alpha] += vi*zI;
			}
		}
		sum += cR[alpha]*cR[alpha] + cI[alpha]*cI[alpha];
	}
	const double wfnorm = 1.0/sqrt(sum);
	for (size_t alpha = 0; alpha < N; ++alpha)
	{
		cR[alpha] *= wfnorm;
		cI[alpha] *= wfnorm;		   
	}
	
	ofstream fgp("movie.gp");
			
	for (double t = 0; t < 36.0; t += 0.0625)
	{
		cout << t << endl;
		for (size_t j = 0; j < N; ++j)
		{
			wfreal[j] = wfimag[j] = 0.0;
			for (size_t alpha = 0; alpha < N; ++alpha)
			{
				double theta = -Eval[alpha]*t;
				while (theta < -M_PI) theta += 2.0*M_PI;
				while (theta > M_PI) theta -= 2.0*M_PI;
				const double costheta = cos(theta);
				const double sintheta = sin(theta);
				const double q = Evec[i0+N*alpha]*Evec[j+N*alpha];
				wfreal[j] += Evec[j+N*alpha]*(cR[alpha]*costheta-cI[alpha]*sintheta);
				wfimag[j] += Evec[j+N*alpha]*(cI[alpha]*costheta+cR[alpha]*sintheta);
			}
		}
		write_mag(wfreal,wfimag,fgp);
	}
	
	fgp.close();
	
	return 0;
}

